//-----------------------------------------------------------------------------
//  Drag & Drop Component
//  Copyright (c) 2003, 04 By Mental Studio -- http://mental.mentsu.com
//  Author: Raptor.z<raptor@mentsu.com>
//  Date: Dec.03-03, Dec.13, Dec.27, Jan.24-04, Feb.20, Dec.04
//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "MOleDragDrop.h"
#pragma package(smart_init)
//-----------------------------------------------------------------------------
//  Initialize & uninitialize OLE environment
//-----------------------------------------------------------------------------

class TOleInitClass
{
public:
    __fastcall TOleInitClass(  ) { OleInitialize( NULL ); }
    __fastcall ~TOleInitClass( ) { OleUninitialize( ); }
};

TOleInitClass oic;

//-----------------------------------------------------------------------------

__fastcall TMOleDropTarget::TMOleDropTarget(TMOleDragDrop* aOwner)
    : FOwner( aOwner ), FRefCount( 1 )
{
    CF_URL  = RegisterClipboardFormat( CFSTR_SHELLURL );
    CF_HTML = RegisterClipboardFormat( "HTML Format" );
    FFormatEtc.cfFormat = CF_TEXT;
    FFormatEtc.ptd      = NULL;
    FFormatEtc.dwAspect = DVASPECT_CONTENT;
    FFormatEtc.lindex   = -1;
    FFormatEtc.tymed    = TYMED_HGLOBAL;
}
//-----------------------------------------------------------------------------

__fastcall TMOleDropTarget::~TMOleDropTarget(void)
{
    FOwner->DropControl = NULL;
}
// -----------------------------------------------------------------------------
//  CAUTION : CF_HTML will be encoded by UTF-8,
//  it must call Utf8ToAnsi to convert to Ansi
//  or call Utf8Decode to convert to Unicode(UCS-2)
//
AnsiString __fastcall TMOleDropTarget::GetDataTextA( IDataObject * aDataObj, tagFORMATETC * aFmtEtc )
{
    STGMEDIUM medium;
    FDragResult = aDataObj->GetData( aFmtEtc, &medium );
    LPSTR p = ( LPSTR )::GlobalLock( medium.hGlobal );
    AnsiString s;
    if ( ( aFmtEtc->cfFormat == CF_UNICODETEXT ) && FOwner->FAnsicode )
        s = Utf8ToAnsi( UTF8Encode( WideString( p ) ) );
    else if ( ( aFmtEtc->cfFormat == CF_HTML ) && FOwner->FAnsicode )
        s = Utf8ToAnsi( AnsiString( p ) );
    else
        s = AnsiString( p );
    ::GlobalUnlock( medium.hGlobal );
    ::ReleaseStgMedium( &medium );
    return s;
}
// -----------------------------------------------------------------------------
//  CAUTION : CF_HTML will be encoded by UTF-8,
//  it must call Utf8ToAnsi to convert to Ansi
//  or call Utf8Decode to convert to Unicode(UCS-2)
//
WideString __fastcall TMOleDropTarget::GetDataTextW( IDataObject * aDataObj, tagFORMATETC * aFmtEtc )
{
    STGMEDIUM medium;
    FDragResult = aDataObj->GetData( aFmtEtc, &medium );
    LPSTR p = ( LPSTR )::GlobalLock( medium.hGlobal );
    WideString s = WideString( p );
    ::GlobalUnlock( medium.hGlobal );
    ::ReleaseStgMedium( &medium );
    return s;
}
// -----------------------------------------------------------------------------
//  CAUTION : File name perhaps is encoded by Unicode(UCS-2),
//  it must call Utf8Encode and Utf8ToAnsi to convert to Ansi
//
int __fastcall TMOleDropTarget::GetDataFiles( IDataObject * aDataObj, tagFORMATETC * aFmtEtc,
    Classes::TStrings* aFiles )
{
    STGMEDIUM medium;
    int i, nCount;
    char sBuf[MAX_PATH];

    FDragResult = aDataObj->GetData( aFmtEtc, &medium );
    try {
        i = -1;
        nCount = ::DragQueryFile( medium.hGlobal, i, sBuf, 0 );  //  Get the Drag _Files Number.
        if ( nCount > 0 )
            for ( i = 0; i < nCount; ++i )
                if ( ::DragQueryFile( medium.hGlobal, i, sBuf, MAX_PATH ) > 0 )
                    aFiles->Add( AnsiString( sBuf ) );
    }
    catch ( ... )
    {
        aFiles->Clear( );
    }
    ::DragFinish( medium.hGlobal );
    ::ReleaseStgMedium( &medium );
    return aFiles->Count;
}
// -----------------------------------------------------------------------------
//	Miscellaneous DIB Function
//
//  From:
// Project:         Drag and Drop Component Suite
// Component Names: TDropBMPSource
// Module:          DropBMPSource
// Description:     Implements Dragging & Dropping of Bitmaps
//                  FROM your application to another.
// Version:         3.7
// Date:            22-JUL-1999
// Target:          Win32, Delphi 3 - Delphi 5, C++ Builder 3, C++ Builder 4
// Authors:         Angus Johnson,   ajohnson@rpi.net.au
//                  Anders Melander, anders@melander.dk
//                                   http://www.melander.dk
// Copyright         1997-99 Angus Johnson & Anders Melander
// -----------------------------------------------------------------------------

void __fastcall TMOleDropTarget::CopyDIBToBitmap( Graphics::TBitmap * aBitmap,
    PBitmapInfo aBitmapInfo, int aDIBSize )
{
    BITMAPFILEHEADER BitmapFileHeader;
    int nFileSize, nInfoSize;

    TMauto_ptr<TMemoryStream> ms( new TMemoryStream( ) );
    // Write DIB to a stream in the BMP file format
    nFileSize = sizeof ( BITMAPFILEHEADER ) + aDIBSize;
    nInfoSize = sizeof ( BITMAPINFOHEADER );
    if ( aBitmapInfo->bmiHeader.biBitCount > 8 )
    {
        if ( aBitmapInfo->bmiHeader.biCompression & BI_BITFIELDS )
            nInfoSize += 12;
    }
    else
        nInfoSize += sizeof ( RGBQUAD ) * ( 1 << ( aBitmapInfo->bmiHeader.biBitCount ) );
    ms->SetSize( nFileSize );
    // Initialize file header
    memset( &BitmapFileHeader, 0, sizeof ( BITMAPFILEHEADER ) );
    BitmapFileHeader.bfType    = 0x4D42; // 'BM' = Windows BMP signature
    BitmapFileHeader.bfSize    = nFileSize; // File size (not needed)
    BitmapFileHeader.bfOffBits = sizeof ( BITMAPFILEHEADER ) + nInfoSize; // Offset of pixel data
    // Save file header
    ms->Write( &BitmapFileHeader, sizeof ( BITMAPFILEHEADER ) );
    // Save TBitmapInfo structure and pixel data
    ms->Write( aBitmapInfo, aDIBSize );
    // Rewind and load bitmap from stream
    ms->Seek( 0, soFromBeginning );
    aBitmap->LoadFromStream( ms );
}
//-----------------------------------------------------------------------------

bool __fastcall TMOleDropTarget::GetDataBitmap( IDataObject * aDataObj, tagFORMATETC * aFmtEtc,
    Graphics::TBitmap * aBitmap )
{
    STGMEDIUM medium, medium2;
    bool nResult = true;

    FDragResult = aDataObj->GetData( aFmtEtc, &medium );
    if ( aFmtEtc->cfFormat == CF_BITMAP )
    {
        aFmtEtc->cfFormat = CF_PALETTE;
        aDataObj->GetData( aFmtEtc, &medium2 );
    }
    else
        memset( &medium2, 0, sizeof ( medium2 ) );
    try {
        if ( medium2.hGlobal )
            aBitmap->LoadFromClipboardFormat( CF_BITMAP, ( UINT )medium.hGlobal, medium2.hGlobal );
        else
        {
            try {
                CopyDIBToBitmap( aBitmap,
                    ( tagBITMAPINFO * )GlobalLock( medium.hGlobal ), GlobalSize( medium.hGlobal ) );
            }
            catch ( ... )
            {
                nResult = false;
            }
            GlobalUnlock( medium.hGlobal );
        }
    }
    catch ( ... )
    {
        nResult = false;
    }
    if ( aFmtEtc->cfFormat == CF_PALETTE )
        ::ReleaseStgMedium( &medium2 );
    ::ReleaseStgMedium( &medium );
    return nResult;
}
//-----------------------------------------------------------------------------

//( Iunkown }
ULONG __stdcall TMOleDropTarget::AddRef( )
{
    InterlockedIncrement( ( long * )&FRefCount );
    return FRefCount;
}
//-----------------------------------------------------------------------------

ULONG __stdcall TMOleDropTarget::Release( )
{
    InterlockedDecrement( ( long * )&FRefCount );
    if ( FRefCount == 0 )
    {
        delete this;
    }
    return FRefCount;
}
//-----------------------------------------------------------------------------

HRESULT __stdcall TMOleDropTarget::QueryInterface( REFIID iid, void ** ppvObject )
{
    if( ( iid == IID_IUnknown ) || ( iid == IID_IDropTarget ) )
    {
        *ppvObject = this;
        AddRef( );
    }
    else
    {
        *ppvObject = NULL;
        return E_NOINTERFACE;
    }
    return S_OK;
}
//-----------------------------------------------------------------------------

//{ IDropTarget }
HRESULT __stdcall TMOleDropTarget::DragEnter( IDataObject * aDataObj,
    DWORD grfKeyState, _POINTL pt, DWORD *pdwEffect )
{
    IEnumFORMATETC * pEnumFE;
    tagFORMATETC fe;
    ULONG nCount;
    TMDropTypes dt;

    FDragResult = E_FAIL;
    if ( FOwner->FEnabledTarget )
    {
        *pdwEffect = DROPEFFECT_NONE;
        aDataObj->EnumFormatEtc( DATADIR_GET, &pEnumFE );
        while ( ( pEnumFE->Next( 1, &fe, &nCount ) == S_OK ) && ( nCount > 0 ) )
        {
            FDragResult = aDataObj->QueryGetData( &fe );
            if ( ( fe.ptd == NULL ) && ( fe.dwAspect == DVASPECT_CONTENT )
                && ( fe.lindex == -1 ) && ( fe.tymed == TYMED_HGLOBAL )
                && ( !FAILED( FDragResult ) ) )
            {
                switch ( fe.cfFormat )
                {
                case CF_TEXT :
                    dt << dtText;
                    break;
                case CF_UNICODETEXT :
                    dt << dtUnicodeText;
                    break;
                case CF_HDROP :
                    dt << dtFiles;
                    break;
                case CF_BITMAP :
                case CF_DIB :
                    dt << dtBitmap;
                    break;
                default :
                    if ( fe.cfFormat == CF_URL )
                        dt << dtUrl;
                    else if ( fe.cfFormat == CF_HTML )
                        dt << dtHtml;
                    break;
                }
            }
        }
        pEnumFE->Release( );
        if ( FOwner->FOnDragEnter )
        {
            fe = FFormatEtc;
            FOwner->FOnDragEnter( FOwner, aDataObj, grfKeyState, pt, pdwEffect, fe, dt );
            if ( *pdwEffect != DROPEFFECT_NONE )
                FFormatEtc = fe;
        }
        else if ( !dt.Empty( ) )
            *pdwEffect = DROPEFFECT_COPY;
        FOwner->FDropTypes = dt;
    }
    FCanDrop = FDragResult;
    return FDragResult;
}
//-----------------------------------------------------------------------------

HRESULT __stdcall TMOleDropTarget::DragOver(DWORD grfKeyState, _POINTL pt,
    DWORD *pdwEffect)
{
    FDragResult = S_OK;
    try {
        if ( FOwner->FOnDragOver )
            FOwner->FOnDragOver( FOwner, grfKeyState, pt, pdwEffect, FOwner->FDropTypes );
    }
    catch ( ... )
    {
        FDragResult = E_FAIL;
    }
    return FDragResult;
}
//-----------------------------------------------------------------------------

HRESULT __stdcall TMOleDropTarget::DragLeave()
{
    FDragResult = S_OK;
    try {
        if ( FOwner->FOnDragLeave )
            FOwner->FOnDragLeave( FOwner );
    }
    catch ( ... )
    {
        FDragResult = E_FAIL;
    }
    return FDragResult;
}
//-----------------------------------------------------------------------------

HRESULT __stdcall TMOleDropTarget::Drop( IDataObject *aDataObj, DWORD grfKeyState,
    _POINTL pt, DWORD *pdwEffect )
{
    FDragResult = S_OK;
    try {
        if ( FOwner->FOnDragDrop )
            FOwner->FOnDragDrop( FOwner, aDataObj, grfKeyState, pt, pdwEffect, FFormatEtc );
        else if ( FOwner->FOnSimpleDrop )
            FOwner->FOnSimpleDrop( FOwner, aDataObj, grfKeyState, pt, pdwEffect, FOwner->FDropTypes );
    }
    catch ( ... )
    {
        FDragResult = E_FAIL;
    }
    return FDragResult;
}
//-----------------------------------------------------------------------------
//  TMOleDragDrop
//-----------------------------------------------------------------------------
// ValidCtrCheck is used to assure that the components created do not have
// any pure virtual functions.
//
//---------------------------------------------------------------------------

static inline void ValidCtrCheck(TMOleDragDrop *)
{
    new TMOleDragDrop(NULL);
}

//---------------------------------------------------------------------------
__fastcall TMOleDragDrop::TMOleDragDrop(TComponent* Owner)
    : TComponent(Owner), FEnabledTarget( true ),
    FAnsicode( true ), FOleDrag( NULL )
{
    if ( !ComponentState.Contains( csDesigning ) )
        FOleDrag = new TMOleDropTarget( this );
}
//---------------------------------------------------------------------------

__fastcall TMOleDragDrop::~TMOleDragDrop(void)
{
    DropControl = NULL;
    if ( FOleDrag )
        delete FOleDrag;
}
//---------------------------------------------------------------------------

void __fastcall TMOleDragDrop::Loaded(void)
{
    Classes::TComponent::Loaded( );
    Owner->FreeNotification( this );
}
//---------------------------------------------------------------------------

void __fastcall TMOleDragDrop::Notification(Classes::TComponent* aComponent, Classes::TOperation aOperation)
{
    if ( FDropControl && ( aOperation == opRemove )
        && ( ( aComponent == FDropControl ) || ( aComponent == Owner ) ) )
        DropControl = NULL;
	Classes::TComponent::Notification( aComponent, aOperation );
}
//---------------------------------------------------------------------------

void __fastcall TMOleDragDrop::SetDropControl(Controls::TWinControl* aValue)
{
    if ( aValue != FDropControl )
    {
        if ( FOleDrag && FDropControl && aValue == NULL )
            try {
                OLECHECK( ::RevokeDragDrop( FDropControl->Handle ) );
            }
            catch ( ... )
            {
            }
        FDropControl = aValue;
        if ( FOleDrag && FDropControl )
            ::RegisterDragDrop( FDropControl->Handle, FOleDrag );
    }
}
//---------------------------------------------------------------------------

void __fastcall TMOleDragDrop::SetEnabledTarget(bool aValue)
{
    if ( aValue != FEnabledTarget )
        FEnabledTarget = aValue;
}
//---------------------------------------------------------------------------

bool __fastcall TMOleDragDrop::GetDropData( IDataObject * aDataObj, TMDropType aDropType, tagFORMATETC * aFmtEtc )
{
    IEnumFORMATETC * pEnumFE;
    ULONG nCount;
    bool nResult = false;

    aDataObj->EnumFormatEtc( DATADIR_GET, &pEnumFE );
    while ( ( pEnumFE->Next( 1, aFmtEtc, &nCount ) == S_OK ) && ( nCount > 0 ) )
    {
        if ( ( ( aFmtEtc->cfFormat == CF_TEXT ) && ( aDropType == dtText ) )
            || ( ( aFmtEtc->cfFormat == CF_UNICODETEXT ) && ( aDropType == dtUnicodeText ) )
            || ( ( aFmtEtc->cfFormat == CF_URL ) && ( aDropType == dtUrl ) )
            || ( ( aFmtEtc->cfFormat == CF_HTML ) && ( aDropType == dtHtml ) )
            || ( ( aFmtEtc->cfFormat == CF_HDROP ) && ( aDropType == dtFiles ) )
            || ( ( ( aFmtEtc->cfFormat == CF_DIB ) || ( aFmtEtc->cfFormat == CF_BITMAP ) )
            && ( aDropType == dtBitmap ) ) )
        {
            nResult = true;
            break;
        }
    }
    pEnumFE->Release( );
    return nResult;
}
//---------------------------------------------------------------------------

bool __fastcall TMOleDragDrop::GetDropData( IDataObject * aDataObj,
    TMDropType aDropType, AnsiString &aText )
{
    tagFORMATETC fe;

    if ( GetDropData( aDataObj, aDropType, &fe ) )
    {
        aText = FOleDrag->GetDataTextA( aDataObj, &fe );
        return true;
    }
    else
        return false;
}
//---------------------------------------------------------------------------

bool __fastcall TMOleDragDrop::GetDropData( IDataObject * aDataObj,
    TMDropType aDropType, WideString &aText )
{
    tagFORMATETC fe;

    if ( GetDropData( aDataObj, aDropType, &fe ) )
    {
        aText = FOleDrag->GetDataTextW( aDataObj, &fe );
        return true;
    }
    else
        return false;
}
//---------------------------------------------------------------------------

bool __fastcall TMOleDragDrop::GetDropData( IDataObject * aDataObj,
    TMDropType aDropType, TStrings * aFiles )
{
    tagFORMATETC fe;

    if ( GetDropData( aDataObj, aDropType, &fe ) )
    {
        FOleDrag->GetDataFiles( aDataObj, &fe, aFiles );
        return true;
    }
    else
        return false;
}
//---------------------------------------------------------------------------

bool __fastcall TMOleDragDrop::GetDropData( IDataObject * aDataObj,
    TMDropType aDropType, TGBitmap * aBitmap )
{
    tagFORMATETC fe;

    if ( GetDropData( aDataObj, aDropType, &fe ) )
    {
        FOleDrag->GetDataBitmap( aDataObj, &fe, aBitmap );
        return true;
    }
    else
        return false;
}
//---------------------------------------------------------------------------

namespace Moledragdrop
{
    void __fastcall PACKAGE Register()
    {
         TComponentClass classes[1] = {__classid(TMOleDragDrop)};
         RegisterComponents("MentalCtrls", classes, 0);
    }

//-----------------------------------------------------------------------------

int CF_URL, CF_HTML;
}
//---------------------------------------------------------------------------

